local AddonName, namespace = ...;

-- upvalues
local _G = _G;
local select = _G.select;

-- check if PitBull4 is installed and running
if select(6, _G.GetAddOnInfo("PitBull4_" .. (debugstack():match("[o%.][d%.][u%.]les\\(.-)\\") or ""))) ~= "MISSING" then return end
local PitBull4 = _G.PitBull4;
if not PitBull4 then error("PitBull4_InlinePower requires PitBull4") end

-- plugin validity check
namespace.invalid = true;

local class = select(2, _G.UnitClass("player")); -- english class name, f.e. PALADIN
local level = _G.UnitLevel("player"); -- character level
local spec = _G.GetSpecialization(); -- the class spec, actually is set directly after hitting the Enter World button

-- Paladins are using Holy Power
if( class == "PALADIN" and _G.PowerBarColor["HOLY_POWER"] ) then
	namespace.invalid = false;
-- Monks are using Chi
elseif( class == "MONK" and _G.PowerBarColor["CHI"] ) then
	namespace.invalid = false;
-- Priests are using Shadow Orbs from level 10 and if specced Shadow, checks later
elseif( class == "PRIEST" ) then
	namespace.invalid = false;
-- Warlocks are using Soul Shards (Affliction), Demon Fury is provided by PitBull itself
-- TODO: check Burning Embers for Destro-Locks, currently my Destro-Lock hasn't Burning Embers as a power
elseif( class == "WARLOCK" and spec == _G.SPEC_WARLOCK_AFFLICTION and _G.PowerBarColor["SOUL_SHARDS"] ) then
	namespace.invalid = false;
-- Death Knights are using Runes
elseif( class == "DEATHKNIGHT" ) then
	namespace.invalid = false;
end

-- if no valid class the script will end
if( namespace.invalid ) then return end
namespace.invalid = nil;

-- if we want the PitBull4 layout to update, we set changed to true (update means to display/hide the power bar)
-- the Layout updater will set it false again after updating
local changed = false;

-- our POWER infotable, used keys are currently:
--  enabled						Whether or not the power bar should be displayed.
--	max								The number of powers the class may use, updated by the addon if it changes due to talents or something else.
--	_max							Static number of powers the class my use, if set, doesn't call UnitPowerMax API. For example: Death Knight runes!
--	name							The name of the power which is used internally by the API. Required for correct updating.
--	spell							The spell id of the power, used by both UnitPower and UnitPowerMax API's.
--  reqlevel					If the power bar needs a specific level to be displayed. 0 means available from the start up.
--	reqspec						If the power is attached to a specific class-spec, for example: Shadow Priest. Used by TALENT_UPDATE event.
--	color							Color table with r, g, b keys. Used as default bar color.
--  color_override		By set a new color table for e.g. index 4, the fourth power will use the new color.
--  events						The events which will be registered when the module gets enabled.
--	updater						The module function (must be a module key) which is responsible for updates. Defaults to module:PowerUpdateHandler

local POWER = {
	enabled = true,
	max = 0,
	reqlevel = 0,
	color_override = {},
	events = {"UNIT_POWER_FREQUENT", "UNIT_DISPLAYPOWER", "PLAYER_ENTERING_WORLD"},
	updater = "PowerUpdateHandler",
};

-- The PitBull4 module itself. Defining it as "bar" will anchor it "inside" the unit frame layout, InlinePower doesn't use a real bar
local module = PitBull4:NewModule("InlinePower", "AceEvent-3.0");
module:SetModuleType("bar");
module:SetName("Inline Power");
module:SetDescription("Replace for PitBulls power icon indicators of spec-specific powers.");
module:SetDefaults({ position = 3 });

-- when the module gets enabled, iterates all POWER events and registers them
function module:OnEnable()
	for _,event in ipairs(POWER.events) do
		self:RegisterEvent(event);
	end
end

-- all events are simply unregistered when disabling the module
function module:OnDisable()
	self:UnregisterAllEvents();
end

-- iterates over all "player" unit PitBull4 frames and updates them
local function update_player(self)
	for frame in PitBull4:IterateFramesForUnitID("player") do
		self:Update(frame);
	end
end

-- PLAYER_ENTERING_WORLD event, may get overridden if needed
function module:PLAYER_ENTERING_WORLD()
	update_player(self);
end

-- UNIT_POWER_FREQUENT event
function module:UNIT_POWER_FREQUENT(event, unit, power)
	if( unit == "player" and power == POWER.name ) then
		update_player(self);
	end
end

-- UNIT_DISPLAYPOWER event, never saw it but all unit frames are listening to it
function module:UNIT_DISPLAYPOWER(event, unit)
	if( unit == "player" ) then
		update_player(self);
	end
end

-- RUNE_POWER_UPDATE event, used only by Death Knight
function module:RUNE_POWER_UPDATE()
	update_player(self);
end

-- RUNE_TYPE_UPDATE event, used only by Death Knight
function module:RUNE_TYPE_UPDATE()
	update_player(self);
end

-- PLAYER_TALENT_UPDATE event, used by some classes
function module:PLAYER_TALENT_UPDATE()
	spec = _G.GetSpecialization();
	
	local new_value = spec == POWER.reqspec;
	if( new_value ~= POWER.enabled ) then
		changed = true;
	end
	
	POWER.enabled = new_value;
	update_player(self);
end

-- fills the POWER table with Paladin specific stuff

if( class == "PALADIN" ) then
	POWER.name = "HOLY_POWER";
	POWER.spell = _G.SPELL_POWER_HOLY_POWER;
	POWER.color = _G.PowerBarColor["HOLY_POWER"];
	
	POWER.reqlevel = _G.PALADINPOWERBAR_SHOW_LEVEL;
	
	POWER.color_override[4] = {r = 0.95, g = 0.2, b = 0.2};
	POWER.color_override[5] = POWER.color_override[4];
-- fills the POWER table with Monk specific stuff	
elseif( class == "MONK" ) then
	POWER.name = "CHI";
	POWER.spell = _G.SPELL_POWER_CHI;
	POWER.color = {r = 0.55, g = 1, b = 0.71}; --_G.PowerBarColor["CHI"]; -- too bright
-- fills the POWER table with Priest specific stuff
elseif( class == "PRIEST" ) then
	POWER.name = "SHADOW_ORBS";
	POWER.spell = _G.SPELL_POWER_SHADOW_ORBS;
	POWER.color = {r = 0.6, g = 0.2, b = 0.8};
	
	POWER.reqspec = _G.SPEC_PRIEST_SHADOW;
	POWER.reqlevel = _G.SHADOW_ORBS_SHOW_LEVEL;
	POWER.enabled = false;
	
	table.insert(POWER.events, "PLAYER_TALENT_UPDATE");
	
	function module:PLAYER_ENTERING_WORLD()
		self:PLAYER_TALENT_UPDATE();
	end
-- fills the POWER table with Warlock specific stuff
elseif( class == "WARLOCK" and spec == _G.SPEC_WARLOCK_AFFLICTION ) then
	POWER.name = "SOUL_SHARDS";
	POWER.spell = _G.SPELL_POWER_SOUL_SHARDS;
	POWER.color = _G.PowerBarColor["SOUL_SHARDS"];
	
	POWER.reqspec = _G.SPEC_WARLOCK_AFFLICTION;
	POWER.enabled = false;
	
	table.insert(POWER.events, "PLAYER_TALENT_UPDATE");
	
	function module:PLAYER_ENTERING_WORLD()
		self:PLAYER_TALENT_UPDATE();
	end
-- fills the POWER table with Death Knight specific stuff
elseif( class == "DEATHKNIGHT" ) then
	POWER._max = 6;
	POWER.name = "RUNES";
	POWER.spell = _G.SPELL_POWER_RUNES;
	POWER.color = {r = 0.8, g = 0.1, b = 1};
	
	POWER.color_override[1] = {r = 0.7, g = 0, b = 0};
	POWER.color_override[2] = POWER.color_override[1];
	POWER.color_override[3] = {r = 0, g = 0.7, b = 0};
	POWER.color_override[4] = POWER.color_override[3];
	POWER.color_override[5] = {r = 0, g = 0.7, b = 0.7};
	POWER.color_override[6] = POWER.color_override[5];
	
	POWER.events = {"PLAYER_ENTERING_WORLD", "RUNE_POWER_UPDATE", "RUNE_TYPE_UPDATE"};
	POWER.updater = "PowerUpdateHandler_DeathKnight";
-- return if a class passed the validity check which isn't handled here
else
	return;
end

-- if the character level is under the required power level, we'll listen for PLAYER_LEVEL_UP events
-- also adds an event listener function to the module
if( level <= POWER.reqlevel ) then
	POWER.enabled = false;
	table.insert(POWER.events, "PLAYER_LEVEL_UP");
	
	function module:PLAYER_LEVEL_UP()
		if( level >= POWER.reqlevel ) then
			self:UnregisterEvent("PLAYER_LEVEL_UP");
			POWER.enabled = true;
			update_player(self);
		end
	end
end

-- values for frame positioning
local FRAME_SPACE = 2;
local FRAME_BORDER = 0;

local last_control_width = 0; -- indicates if we need to resize the frames
local last_control_height = 0; -- indicates if we need to resize the frames

-- the Update function is called by either PitBull4 or us
function module:UpdateFrame(frame)
	-- additional powers are player-specific, we cannot display other players extra powers
	-- also removes the power bar if we set POWER.enabled to false
	if( frame.unit ~= "player" or not POWER.enabled ) then
		return self:ClearFrame(frame);
	end
	
	local id = self.id; -- the module name, InlinePower
	local control = frame[id]; -- our power bar is stored in the PitBull4 frame and accessible by the module name
	local made_control = not control;
	
	if( made_control ) then
		-- we need to create a BetterStatusBar because the module is registered as "bar", so PitBull4 trys to access bar-specific functions
		-- the bar itself is only used as a container for anchoring our little frames
		-- I know its dirty but I didn't figure out another way of getting inner unit frame objects, even the "Blank Space" module is a bar
		control = PitBull4.Controls.MakeBetterStatusBar(frame);
		frame[id] = control; -- save it to the PitBull4 frame
	end
	
	-- current maximum power, always 6 for death knights
	local power_max;
	if( POWER._max ) then
		power_max = POWER._max;
	else
		power_max = _G.UnitPowerMax("player", POWER.spell)
	end
	
	POWER.current = _G.UnitPower("player", POWER.spell); -- current power value
	
	local control_width = control:GetWidth();
	local control_height = control:GetHeight();
	
	if( control_width ~= last_control_width or control_height ~= last_control_height or power_max ~= POWER.max ) then
		last_control_width = control_width;
		last_control_height = control_height;
		
		local width = (control_width - ((power_max - 1) * FRAME_SPACE)) / power_max; -- calculates the width for each power object
		local half = width / 2; -- and its half
		
		local power;
		for i = 1, power_max do
			if( not control[i] ) then
				control[i] = PitBull4.Controls.MakeInlinePower(control, i);
			end
			power = control[i];
			
			if( not power:IsShown() ) then
				power:Show();
			end
			power:ClearAllPoints();
			power:SetSize(width, control_height);
			power:SetPoint("CENTER", control, "LEFT", FRAME_BORDER + (i - 1) * (FRAME_SPACE + width) + half, 0);
		end
		
		-- we need to hide all power displays which are currently not used
		if( power_max > 0 and power_max < POWER.max ) then
			for i = power_max, POWER.max do
				if( control[i] ) then
					control[i]:Hide();
				end
			end
		end
		POWER.max = power_max;
	end
	
	if( not control:IsVisible() ) then
		control:Show();
	end
	
	module[POWER.updater](module, control);
	
	made_control = changed or made_control;
	changed = false;
	
	return made_control;
end

function module:PowerUpdateHandler(control)
	for i = 1, POWER.max do
		control[i]:UpdateColor(POWER.color_override[i] or POWER.color);
		
		if( i <= POWER.current ) then
			control[i]:Activate();
		else
			control[i]:Deactivate();
		end
	end
end

function module:PowerUpdateHandler_DeathKnight(control)
	for i = 1, POWER.max do
		control[i]:UpdateColor(_G.GetRuneType(i) == 4 and POWER.color or POWER.color_override[i]);
		
		local start, duration, ready = _G.GetRuneCooldown(i);
		
		if( ready ) then
			control[i]:Activate();
		else
			control[i]:Deactivate();
		end
	end
end

-- bar functions which need to be set, but they always return 0
function module:GetValue(frame) return 0 end
module.GetExampleValue = module.GetValue;
function module:GetColor(frame, value) return 0, 0, 0, 0 end
module.GetExampleColor = module.GetColor;
